#pragma once

//Streams
#include <iostream>
#include <fstream>

//For random numbers generator
#include <cstdlib>
#include <ctime>
#include <random>

#include <algorithm>
#include <iterator>
#include <ranges>
#include <numeric>

#include "Exception.h"
#include "ContainerParsing.cpp"

//Sequence containers
#include <vector>
#include <list>
#include <deque>


//=================================================================

void showSequenceContainersMenu()
{
	std::cout << std::endl << "Work with sequence containers." << std::endl;
	std::cout << "Choose what to do:" << std::endl;
	std::cout << "1. Create data in file and load it into container." << std::endl;
	std::cout << "2. Create data in file with instant record into container and modify data." << std::endl;
	std::cout << "3. Load data from file into container" << std::endl;
	std::cout << "4. Load data from file to given container and modify data(+adding of average of all elements to the end of container)." << std::endl;
	std::cout << "5. Modify data in file." << std::endl;
	std::cout << "6. Output data from container into given file." << std::endl;
	std::cout << "7. Output data from container into temporary file." << std::endl;
	std::cout << "8. Quit" << std::endl << std::endl;
}


//=================================================================


//Creates text file with 100 random integers in range [-50, 50]
void createData(const char* filename)
{
	std::fstream file(filename, std::ios::out | std::ios::trunc);


	std::srand(unsigned(std::time(nullptr)));

	if (!file.is_open())
	{
		throw MyException("Failure. File was not opened...\n");
	}
	else
	{
		for (int i : std::ranges::views::iota(1, 101))
		{
			file << std::rand() % 101 - 50 << " ";
			if ((i % 20 == 0) && (i != 0)) file << "\n";
		}

		file.close();
	}
}


//Creates text file with 100 random integers in range [-50, 50] with recording in container
template<typename _ContType>
void createData(const char* filename, _ContType& cont)
{
	std::fstream file(filename, std::ios::out | std::ios::trunc);
    //     
	std::ostream_iterator<_ContType::value_type> file_iter(file, " ");

	if (!file.is_open())
	{
		throw MyException("Failure. File was not opened...\n");
	}
	else
	{
		srand((unsigned)time(nullptr));
		cont.clear();
		for (int i : std::ranges::views::iota(1, 101))
		{
			*std::back_inserter(cont) = std::rand() % 101 - 50;
		}

		std::ranges::copy(cont, file_iter);

		file.close();
	}
}


//=================================================================


//Reads data from file into sequence container(usage of file streams)
template<typename _ContType>
_ContType loadData(const char* filename)
{
	std::ifstream file(filename, std::ios::in);

	if (!file.is_open())
	{
		throw MyException("Failure.File was not opened...\n");
	}
	else
	{
		_ContType cont;
		std::string cur_num;
		file.seekg(0);
		while (!file.eof())
		{
			cur_num[0] = '\n';
            //   
			file >> cur_num;
			*std::back_inserter(cont) = atoi(cur_num.c_str());
		}

		cont.pop_back(); //      

		file.close();

		return cont;
	}
}


//Reads data from file into given sequence container(usage of stream iterators)
template<typename _ContType>
void loadData(const char* filename, _ContType& cont)
{
	std::ifstream file(filename, std::ios::in);

	if (!file.is_open())
	{
		throw MyException("Failure.File was not opened...\n");
	}
	else
	{
		cont.clear();
		file.seekg(0); //rewind
        //   
		std::copy(std::istream_iterator<_ContType::value_type>(file),
				  std::istream_iterator<_ContType::value_type>(),
				  std::back_inserter(cont));

		file.close();
	}

}


//=================================================================


//Modifying of SEQUENCE container(multiplying of every container element on minimal element in container
//and adding of average of all container elements to the end of container)
template<typename _ContType>
_ContType::value_type modify(_ContType& cont)
{
	if (cont.size() == 0)
		throw MyException("\nProgram cannot modify elements of an empty container!\n");

	typename _ContType::value_type min_el = *std::min_element(std::begin(cont), std::end(cont), std::less<_ContType::value_type>()); //minimal element in container
	std::for_each(cont.begin(), cont.end(), [=](_ContType::reference cur_container_el) {cur_container_el *= min_el; });
	typename _ContType::value_type sum = std::accumulate(std::begin(cont), std::end(cont), 0, std::plus<_ContType::value_type>());
	cont.push_back(std::fabs(sum * 1.0 / cont.size()));

	return min_el;
}


//Modifying of range contained in SEQUENCE container(multiplying of every range elements on minimal element in this range)
template<typename _IterType>
 auto modify(_IterType beg, _IterType end) -> std::iterator_traits<_IterType>::value_type
{
	 //???
	if (std::distance(beg, end) <= 0)
		throw MyException("Iterators of diapazon are not in valid sequence!");

	typedef typename std::iterator_traits<_IterType>::value_type value_type; //definition of type na kotoriy ukazivaet iterator with type "_IterType"

	auto min_el = *std::ranges::min_element(beg, end, std::less<value_type>()); //minimal element in container
	std::ranges::for_each(beg, end, [=](value_type& cur_el) {cur_el *= min_el; });

	return min_el;
}

 
 template<typename dataType>
const char* modify(const char* filename)
{
	const char* tempFileName = "temp.txt";
	std::ifstream file(filename);
	std::ofstream res_file(tempFileName, std::ios::trunc);
	std::ostream_iterator<dataType> res_file_it(res_file, " ");


	if ((!file.is_open()) || (!res_file.is_open()))
		throw MyException("Failed to load file for modification");
	
	auto min_el = std::min_element(std::istream_iterator<dataType>(file), std::istream_iterator<dataType>(), std::less<dataType>());
	file.close();
	file.open(filename);
	std::transform(std::istream_iterator<dataType>(file), std::istream_iterator<dataType>(),
				   res_file_it,
				   [min_el](auto cur_num) {return cur_num * (*min_el); });

	file.close();
	res_file.close();
	std::remove(filename);
	std::rename(tempFileName, filename);

	return filename;
}

//=================================================================


//
template<typename _ContType>
void outputResult(const char* filename, const _ContType& cont)
{
	std::ofstream outp_file(filename, std::ios::out | std::ios::trunc);
	int cnt = 0; //amount of processed elements


	if (!outp_file.is_open())
		throw MyException("Error! Failure in opening output file.");

	outp_file.seekp(0);

	for (auto cur_el : cont)
	{
		outp_file << cur_el << " ";
		if ((cnt > 0) && (cnt % 20 == 0))
			outp_file << "\n";
		++cnt;
	}

	outp_file.close();
}


template<typename _ContType>
const char* outputResult(const _ContType& cont)
{
	const char* resFileName = "Result.txt";
	std::ofstream outp_file(resFileName, std::ios::out | std::ios::trunc);
	std::ostream_iterator<_ContType::value_type> file_iter(outp_file, " ");


	if (!outp_file.is_open())
		throw MyException("Error! Failure in opening output file.");

	std::copy(cont.begin(), cont.end(), file_iter);

	outp_file.close();

	return resFileName;
}


void workWithSequenceContainers()
{
	int key = 0;
	const char* filename = "100_Random_Nums.txt";
	std::list<int> data_list;
	std::remove(filename);

	while (key != 8)
	{
		try
		{
			std::cout << "\n\n";
			showSequenceContainersMenu();
			std::cout << "-> "; std::cin >> key;

			switch (key)
			{
			case 1:
			{
				system("cls");
				std::cout << "1. Create data in file and load it into container." << std::endl;
				createData(filename);
				std::vector<int> data_vec = loadData<std::vector<int>>(filename);
				//data_list.assign(data_vec.begin(), data_vec.end()); //,  move???
				std::move(data_vec.begin(), data_vec.end(), std::back_inserter(data_list));
				parseContWithNums(data_list); 
				break;
			}
			case 2:
			{
				system("cls");
				std::cout << "2. Create data in file with instant record into container and modify data." << std::endl;
				
				createData(filename, data_list);
				parseContWithNums(data_list);
				auto min = modify(data_list.begin(), data_list.end());
				std::cout << "\n\nMinimal element in container: " << min << "\n\n";
				std::cout << "After modification...\n";
				parseContWithNums(data_list);
				break;
			}
			case 3:
			{
				system("cls");
				std::cout << "3. Load data from file(into container) and show." << std::endl;
				loadData(filename, data_list);
				parseContWithNums(data_list);
				break;
			}
			case 4:
			{
				system("cls");
				std::cout << "4. Load data from file to given container and modify data(+adding of average of all elements to the end of container)." << std::endl;
				loadData(filename, data_list);
				parseContWithNums(data_list);
				auto min = modify(data_list);
				std::cout << "\n\nMinimal element in container: " << min << "\n\n";
				std::cout << "After modification...\n";
				parseContWithNums(data_list);
				break;
			}
			case 5:
			{
				system("cls");
				std::cout << "5. Modify data in file." << std::endl;
				modify<int>(filename);
				std::cout << "Data was successfully modified.\n\n";
				break;
			}
			case 6:
			{
				system("cls");
				std::cout << "6. Output data from container into given file." << std::endl;
				if (data_list.empty())
				{
					std::cout << "Data absents in container!\n";
				}
				else
				{
					outputResult(filename, data_list);
					std::cout << "Data was successfully outputed into file.\n";
				}
				break;
			}
			case 7:
			{
				system("cls");
				std::cout << "7. Output data from container into temporary file." << std::endl;
				if (data_list.empty())
				{
					std::cout << "Data absents in container!\n";
				}
				else
				{
					const char* res_file = outputResult(data_list);
					std::cout << "Data was successfully outputed into temporary file.\n";
					std::vector<int> temp_data_vec = loadData<std::vector<int>>(res_file);
					std::cout << "Content of temporary file was loaded into temporary container. Here it is:\n";
					parseContWithNums(temp_data_vec);
				}
				break;
			}
			case 8:
			{
				system("cls");
				break;
			}
			default:
			{
				std::cout << "\nWrong input!\n";
			}
			}
		}
		catch (MyException exc)
		{
			std::cout << exc.what();
		}
	}
	
	std::remove(filename);
}


/*
template<typename _DataType>
std::deque<_DataType> loadData(std::fstream& file)
{
	std::deque<_DataType> tmp_deque;
	_DataType cur_elem;

	file.seekp(0, std::ios::beg);
	while (file >> cur_elem)
	{
		if (cur_elem % 2 == 0)
		{
			tmp_deque.push_back(cur_elem);
		}
		else
		{
			tmp_deque.push_front(cur_elem);
		}
	}

	return tmp_deque;
}
*/

/*
//Reads data from file into deque with given condition
template<typename _DataType, typename _Pred>
std::deque<_DataType> loadData(std::fstream& file, _Pred cond)
{
	std::deque<_DataType> tmp_deque;
	_DataType cur_elem;

	file.seekp(0, std::ios::beg);
	while (file >> cur_elem)
	{
		if (cond(cur_elem))
		{
			tmp_deque.push_back(cur_elem);
		}
		else
		{
			tmp_deque.push_front(cur_elem);
		}
	}

	return tmp_deque;
}
*/


